from cpython cimport PyBytes_AsString
from struct import pack


cdef bytes HEADER = b"PGCOPY\n\xff\r\n\x00\x00\x00\x00\x00\x00\x00\x00\x00"
cdef bytes FINALIZE = b"\xff\xff"
cdef bytes NULLABLE = b"\xff\xff\xff\xff"


cpdef long read_num_columns(
    object fileobj,
    long column_length,
):
    """Read one record to bytes."""

    cdef bytes _bytes = fileobj.read(column_length)
    cdef const unsigned char *buf = <const unsigned char*>PyBytes_AsString(
        _bytes
    )
    return (buf[0] << 8) | buf[1]


cpdef object read_record(
    object fileobj,
    object reader,
    object pgoid_function,
    object buffer_object,
    long pgoid,
):
    """Read one record to bytes."""

    cdef bytes _bytes = fileobj.read(4)
    cdef const unsigned char *buf = <const unsigned char*>PyBytes_AsString(
        _bytes
    )
    cdef int length = (buf[0] << 24) | (buf[1] << 16) | (buf[2] << 8) | buf[3]

    if length == 0xffffffff:
        return

    return reader(fileobj.read(length), pgoid_function, buffer_object, pgoid)


cpdef long long skip_all(
    object fileobj,
    long column_length,
    long num_columns,
    long long num_rows,
):
    """Skip all records."""

    cdef long columns = num_columns

    while columns != 0xffff:

        for _ in range(num_columns):
            skip_record(fileobj)

        num_rows += 1
        columns = read_num_columns(fileobj, column_length)

    return num_rows


cdef void skip_record(object fileobj):
    """Skip one record."""

    cdef bytes _bytes = fileobj.read(4)
    cdef const unsigned char *buf = <const unsigned char*>PyBytes_AsString(
        _bytes
    )
    cdef int length = (buf[0] << 24) | (buf[1] << 16) | (buf[2] << 8) | buf[3]
    fileobj.read(length)


cpdef long long writer(
    object fileobj,
    object write_row,
    object dtype_values,
    long num_columns,
):
    """Write pgcopy into fileobj."""

    cdef long long pos = 0

    for buffer_object in make_rows(write_row, dtype_values, num_columns):
        pos += fileobj.write(buffer_object)

    return pos


cpdef bytes nullable_writer(
    object write_dtype,
    object dtype_value,
    object pgoid_function,
    object buffer_object,
    long pgoid,
):
    """Function for write None value and data with length."""

    if dtype_value is None:
        return NULLABLE

    cdef bytes binary_data = write_dtype(
        dtype_value,
        pgoid_function,
        buffer_object,
        pgoid,
    )
    cdef int len_data = len(binary_data)
    return pack(f"!I{len_data}s", len_data, binary_data)


def make_rows(object write_row, object dtype_values, long num_columns):
    """Make pgcopy rows."""

    cdef list bytes_buffer = [HEADER]
    cdef bytes num_columns_bytes = bytes([
        (num_columns >> 8) & 0xFF,
        num_columns & 0xFF
    ])

    for dtype_value in dtype_values:
        bytes_buffer.append(num_columns_bytes)

        for row in write_row(dtype_value):
            bytes_buffer.append(row)

        if len(bytes_buffer) > 1024:
            yield b"".join(bytes_buffer)
            bytes_buffer.clear()

    bytes_buffer.append(FINALIZE)
    yield b"".join(bytes_buffer)



from typing import Any


def read_array(
    binary_data: bytes,
    pgoid_function: object,
    buffer_object: object,
    pgoid: int,
) -> list[Any]:
    """Unpack array values."""

    ...


def write_array(
    dtype_value: list[Any],
    pgoid_function: object,
    buffer_object: object,
    pgoid: int,
) -> bytes:
    """Pack array values."""

    ...



from io import (
    BufferedReader,
    BytesIO,
)
from typing import (
    Any,
    Generator,
    TYPE_CHECKING,
)
from struct import unpack

from .common import (
    ArrayOidToOid,
    PGOid,
    PGOidToDType,
    PGCopySignatureError,
)
from .common.base import (
    read_num_columns,
    read_record,
    skip_all,
)
from .common.dtypes import PostgreSQLDtype

if TYPE_CHECKING:
    from types import FunctionType


class PGCopyReader:
    """PGCopy dump reader."""

    def __init__(
        self,
        file: BufferedReader,
        pgtypes: list[PGOid] = [],
    ) -> None:
        """Class initialization."""

        self.file = file
        self.pgtypes = pgtypes

        self.header: bytes = self.file.read(11)

        if self.header != b"PGCOPY\n\xff\r\n\x00":
            msg = "PGCopy signature not match!"
            raise PGCopySignatureError(msg)

        self.flags_area: list[int] = [
            (byte >> i) & 1
            for byte in self.file.read(4)
            for i in range(7, -1, -1)
        ]
        self.is_oid_enable: bool = bool(self.flags_area[16])
        self.header_ext_length: int = unpack("!i", self.file.read(4))[0]

        if self.is_oid_enable:
            self.column_length: int = 6
        else:
            self.column_length: int = 2

        self.num_columns: int = read_num_columns(
            self.file,
            self.column_length,
        )
        self.num_rows: int = 0
        self.postgres_dtype: list[PostgreSQLDtype] = [
            PGOidToDType[self.pgtypes[column]]
            if self.pgtypes else PostgreSQLDtype.Bytes
            for column in range(self.num_columns)
        ]
        self.pgoid_functions: list[FunctionType] = [
            PGOidToDType[ArrayOidToOid[self.pgtypes[column]]].read
            if self.pgtypes and ArrayOidToOid.get(
                self.pgtypes[column]
            ) else None
            for column in range(self.num_columns)
        ]
        self.pgoid: list[int] = [
            ArrayOidToOid[self.pgtypes[column]].value
            if self.pgtypes and ArrayOidToOid.get(
                self.pgtypes[column]
            ) else 0
            for column in range(self.num_columns)
        ]
        self.buffer_object = BytesIO()

    def __count_rows(self) -> None:
        """Count rows."""

        if not self.num_rows:
            self.num_rows = skip_all(
                self.file,
                self.column_length,
                self.num_columns,
                self.num_rows,
            )

    def read_row(self) -> Generator[Any, None, None]:
        """Read single row."""

        for postgres_dtype, pgoid_function, pgoid in zip(
            self.postgres_dtype,
            self.pgoid_functions,
            self.pgoid,
        ):
            yield read_record(
                self.file,
                postgres_dtype.read,
                pgoid_function,
                self.buffer_object,
                pgoid,
            )

    def to_rows(self) -> Generator[list[Any], None, None]:
        """Read all rows."""

        columns = self.num_columns

        while columns != 0xffff:
            yield [*self.read_row()]
            self.num_rows += 1
            columns = read_num_columns(
                self.file,
                self.column_length,
            )

    def tell(self) -> int:
        """Return current position."""

        return self.file.tell()

    def close(self) -> None:
        """Close file object."""

        self.file.close()

    def __repr__(self) -> str:
        """PGCopy info in interpreter."""

        return self.__str__()

    def __str__(self) -> str:
        """PGCopy info."""

        self.__count_rows()

        if not self.pgtypes:
            self.pgtypes = [
                PGOid.bytea
                for _ in range(self.num_columns)
            ]

        def to_col(text: str) -> str:
            """Format string element."""

            text = text[:14] + "…" if len(text) > 15 else text
            return f" {text: <15} "

        empty_line = (
            "│-----------------+-----------------│"
        )
        end_line = (
            "└─────────────────┴─────────────────┘"
        )
        _str = [
            "<PGCopy dump reader>",
            "┌─────────────────┬─────────────────┐",
            "│ Column Number   │ PostgreSQL Type │",
            "╞═════════════════╪═════════════════╡",
        ]

        for column, pgtype in zip(range(self.num_columns), self.pgtypes):
            _str.append(
                f"│{to_col(f'Column_{column}')}│{to_col(pgtype.name)}│",
            )
            _str.append(empty_line)

        _str[-1] = end_line
        return "\n".join(_str) + f"""
Total columns: {self.num_columns}
Total rows: {self.num_rows}
"""



from io import (
    BytesIO,
    BufferedWriter,
)
from typing import (
    Any,
    Generator,
    TYPE_CHECKING,
)

from .common import (
    ArrayOidToOid,
    PGOid,
    PGCopyRecordError,
    PGOidToDType,
)
from .common.base import (
    make_rows,
    nullable_writer,
    writer,
)

if TYPE_CHECKING:
    from types import FunctionType
    from .common.dtypes import PostgreSQLDtype


class PGCopyWriter:
    """PGCopy dump writer."""

    def __init__(
        self,
        file: BufferedWriter | None,
        pgtypes: list[PGOid],
    ) -> None:
        """Class initialization."""

        if not pgtypes:
            raise PGCopyRecordError("PGOids not defined!")

        self.file = file
        self.pgtypes = pgtypes
        self.num_columns = len(pgtypes)
        self.num_rows = 0
        self.pos = 0
        self.postgres_dtype: list[PostgreSQLDtype] = [
            PGOidToDType[pgtype]
            for pgtype in pgtypes
        ]
        self.pgoid_functions: list[FunctionType] = [
            PGOidToDType[ArrayOidToOid[self.pgtypes[column]]].write
            if self.pgtypes and ArrayOidToOid.get(
                self.pgtypes[column]
            ) else None
            for column in range(self.num_columns)
        ]
        self.pgoid: list[int] = [
            ArrayOidToOid[self.pgtypes[column]].value
            if self.pgtypes and ArrayOidToOid.get(
                self.pgtypes[column]
            ) else 0
            for column in range(self.num_columns)
        ]
        self.buffer_object = BytesIO()

    def write_row(self, dtype_values: Any) -> Generator[Any, None, None]:
        """Write single row."""

        for postgres_dtype, dtype_value, pgoid_function, pgoid in zip(
            self.postgres_dtype,
            dtype_values,
            self.pgoid_functions,
            self.pgoid,
        ):
            yield nullable_writer(
                postgres_dtype.write,
                dtype_value,
                pgoid_function,
                self.buffer_object,
                pgoid,
            )
        self.num_rows += 1

    def from_rows(self, dtype_values: list[Any]) -> Generator[
        bytes,
        None,
        None,
    ]:
        """Write all rows."""

        return make_rows(self.write_row, dtype_values, self.num_columns)

    def write(self, dtype_values: list[Any]) -> None:
        """Write all rows into file."""

        if self.file is None:
            raise PGCopyRecordError("File not defined!")

        self.pos = writer(
            self.file,
            self.write_row,
            dtype_values,
            self.num_columns,
        )

    def tell(self) -> int:
        """Return current position."""

        return self.pos

    def close(self) -> None:
        """Close file object."""

        if self.file:
            self.file.close()

    def __repr__(self) -> str:
        """PGCopy info in interpreter."""

        return self.__str__()

    def __str__(self) -> str:
        """PGCopy info."""

        def to_col(text: str) -> str:
            """Format string element."""

            text = text[:14] + "…" if len(text) > 15 else text
            return f" {text: <15} "

        empty_line = (
            "│-----------------+-----------------│"
        )
        end_line = (
            "└─────────────────┴─────────────────┘"
        )
        _str = [
            "<PGCopy dump writer>",
            "┌─────────────────┬─────────────────┐",
            "│ Column Number   │ PostgreSQL Type │",
            "╞═════════════════╪═════════════════╡",
        ]

        for column, pgtype in zip(range(self.num_columns), self.pgtypes):
            _str.append(
                f"│{to_col(f'Column_{column}')}│{to_col(pgtype.name)}│",
            )
            _str.append(empty_line)

        _str[-1] = end_line

        return "\n".join(_str) + f"""
Total columns: {self.num_columns}
Total rows: {self.num_rows}
"""



def read_point(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> tuple[float, float]:
    """Unpack point value."""

    ...


def write_point(
    dtype_value: tuple[float, float],
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack point value."""

    ...


def read_line(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> tuple[float, float, float]:
    """Unpack line value."""

    ...


def write_line(
    dtype_value: tuple[float, float, float],
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack line value."""

    ...


def read_lseg(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> list[tuple[float, float], tuple[float, float]]:
    """Unpack lseg value."""

    ...


def write_lseg(
    dtype_value: list[tuple[float, float], tuple[float, float]],
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack lseg value."""

    ...


def read_box(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> tuple[tuple[float, float], tuple[float, float]]:
    """Unpack box value."""

    ...


def write_box(
    dtype_value: tuple[tuple[float, float], tuple[float, float]],
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack box value."""

    ...


def read_path(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> list[tuple[float, ...]] | tuple[tuple[float, ...]]:
    """Unpack path value."""

    ...


def write_path(
    dtype_value: list[tuple[float, ...]] | tuple[tuple[float, ...]],
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack path value."""

    ...


def read_polygon(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> tuple[float, ...]:
    """Unpack polygon value."""

    ...


def write_polygon(
    dtype_value: tuple[float, ...],
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack polygon value."""

    ...



from decimal import Decimal


def read_bool(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bool:
    """Unpack bool value."""

    ...


def write_bool(
    dtype_value: bool,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack bool value."""

    ...


def read_oid(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> int:
    """Unpack oid value."""

    ...


def write_oid(
    dtype_value: int,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack oid value."""

    ...


def read_serial2(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> int:
    """Unpack serial2 value."""

    ...


def write_serial2(
    dtype_value: int,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack serial2 value."""

    ...


def read_serial4(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> int:
    """Unpack serial4 value."""

    ...


def write_serial4(
    dtype_value: int,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack serial4 value."""

    ...


def read_serial8(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> int:
    """Unpack serial8 value."""

    ...


def write_serial8(
    dtype_value: int,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack serial8 value."""

    ...


def read_int2(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> int:
    """Unpack int2 value."""

    ...


def write_int2(
    dtype_value: int,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack int2 value."""

    ...


def read_int4(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> int:
    """Unpack int4 value."""

    ...


def write_int4(
    dtype_value: int,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack int4 value."""

    ...


def read_int8(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> int:
    """Unpack int8 value."""

    ...


def write_int8(
    dtype_value: int,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack int8 value."""

    ...


def read_money(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> float:
    """Unpack money value."""

    ...


def write_money(
    dtype_value: float,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack money value."""

    ...


def read_float4(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> float:
    """Unpack float4 value."""

    ...


def write_float4(
    dtype_value: float,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack float4 value."""

    ...


def read_float8(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> float:
    """Unpack float8 value."""

    ...


def write_float8(
    dtype_value: float,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack float8 value."""

    ...


def read_numeric(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> Decimal:
    """Unpack numeric value."""

    ...


def write_numeric(
    dtype_value: Decimal,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack numeric value."""

    ...



cpdef list read_array(
    bytes binary_data,
    object pgoid_function,
    object buffer_object,
    long pgoid,
)
cpdef bytes write_array(
    list dtype_value,
    object pgoid_function,
    object buffer_object,
    long pgoid,
)



cpdef str read_text(
    bytes binary_data,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes write_text(
    str dtype_value,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef str read_macaddr(
    bytes binary_data,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes write_macaddr(
    str dtype_value,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef str read_bits(
    bytes binary_data,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes write_bits(
    str dtype_value,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes read_bytea(
    bytes binary_data,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes write_bytea(
    bytes dtype_value,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)



from decimal import (
    ROUND_HALF_UP,
    Decimal,
)
from libc.math cimport round
from struct import (
    pack,
    unpack,
    unpack_from,
)


cpdef object read_bool(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack bool value."""

    return unpack("!?", binary_data)[0]


cpdef bytes write_bool(
    object dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack bool value."""

    return pack("!?", bool(dtype_value))


cpdef unsigned long read_oid(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack oid value."""

    return unpack("!I", binary_data)[0]


cpdef bytes write_oid(
    object dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack oid value."""

    return pack("!I", <unsigned long>dtype_value)


cpdef unsigned short read_serial2(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack serial2 value."""

    return unpack("!H", binary_data)[0]


cpdef bytes write_serial2(
    object dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack serial2 value."""

    return pack("!H", <unsigned short>dtype_value)


cpdef unsigned long read_serial4(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack serial4 value."""

    return unpack("!L", binary_data)[0]


cpdef bytes write_serial4(
    object dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack serial4 value."""

    return pack("!L", <unsigned long>dtype_value)


cpdef unsigned long long read_serial8(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack serial8 value."""

    return unpack("!Q", binary_data)[0]


cpdef bytes write_serial8(
    object dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack serial8 value."""

    return pack("!Q", <unsigned long long>dtype_value)


cpdef short read_int2(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack int2 value."""

    return unpack("!h", binary_data)[0]


cpdef bytes write_int2(
    object dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack int2 value."""

    return pack("!h", <short>dtype_value)


cpdef long read_int4(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack int4 value."""

    return unpack("!l", binary_data)[0]


cpdef bytes write_int4(
    object dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack int4 value."""

    return pack("!l", <long>dtype_value)


cpdef long long read_int8(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack int8 value."""

    return unpack("!q", binary_data)[0]


cpdef bytes write_int8(
    object dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack int8 value."""

    return pack("!q", <long long>dtype_value)


cpdef double read_money(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack money value."""

    return read_int8(binary_data) * 0.01


cpdef bytes write_money(
    object dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack money value."""

    return write_int8(<long long>round(dtype_value / 0.01))


cpdef float read_float4(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack float4 value."""

    return unpack("!f", binary_data)[0]


cpdef bytes write_float4(
    object dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack float4 value."""

    return pack("!f", <float>dtype_value)


cpdef double read_float8(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack float8 value."""

    return unpack("!d", binary_data)[0]


cpdef bytes write_float8(
    object dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack float8 value."""

    return pack("!d", <double>dtype_value)


cpdef object read_numeric(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack numeric value."""

    cdef:
        int ndigits, weight, sign, dscale
        int i, pos
        bint is_negative
        list digits = []
        object numeric, scale, power, term
        short digit

    ndigits, weight, sign, dscale = unpack_from("!hhhh", binary_data)

    if sign == 0xc000:
        return Decimal("nan")

    is_negative = (sign == 0x4000)

    cdef int data_len = len(binary_data)

    for i in range(8, 8 + ndigits * 2, 2):
        if i + 1 < data_len:
            digit = unpack_from("!h", binary_data, i)[0]
            digits.append(digit)

    numeric = Decimal(0)
    scale = Decimal(10) ** -dscale

    cdef object weight_dec = Decimal(weight)
    cdef object ten = Decimal(10)
    cdef object four = Decimal(4)

    for pos, digit in enumerate(digits):
        power = four * (weight_dec - Decimal(pos))
        term = Decimal(digit) * (ten ** power)
        numeric += term

    if is_negative:
        numeric *= -1

    return numeric.quantize(scale)


cpdef bytes write_numeric(
    object dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack numeric value."""

    cdef bint is_negative
    cdef int sign, dscale, ndigits, weight, digit
    cdef object abs_value, scaled_value, int_value
    cdef list digits = []
    cdef list digit_bytes_list = []
    cdef bytes header, digits_data

    dtype_value = Decimal(dtype_value)

    if dtype_value.is_nan():
        return pack("!hhhh", 0, 0, 0xc000, 0)

    is_negative = dtype_value < 0
    sign = 0x4000 if is_negative else 0x0000

    if dtype_value == 0:
        return pack("!hhhh", 0, 0, sign, 0)

    abs_value = abs(dtype_value)
    as_tuple = abs_value.as_tuple()
    dscale = abs(as_tuple.exponent) if as_tuple.exponent < 0 else 0
    scaled_value = abs_value * (Decimal(10) ** dscale)
    int_value = int(scaled_value.to_integral_value(rounding=ROUND_HALF_UP))
    temp = int_value
    base = 10000

    while temp > 0:
        digits.append(temp % base)
        temp //= base

    if not digits:
        digits = [0]
    else:
        digits.reverse()

    ndigits = len(digits)

    if int_value == 0:
        weight = 0
    else:
        integer_digits = len(str(int(abs_value)))
        weight = (integer_digits - 1) // 4

    header = pack("!hhhh", ndigits, weight, sign, dscale)

    for digit in digits:
        digit_bytes_list.append(pack("!h", digit))

    digits_data = b''.join(digit_bytes_list)

    return header + digits_data



"""Convert between bytes and data types functions."""

from datetime import (
    date,
    datetime as dt,
    time,
)
from dateutil.relativedelta import relativedelta as rd
from enum import Enum
from ipaddress import (
    IPv4Address as IP4A,
    IPv4Network as IP4N,
    IPv6Address as IP6A,
    IPv6Network as IP6N,
)
from typing import NamedTuple
from types import FunctionType

from .functions import (
    read_array,
    read_bits,
    read_bool,
    read_box,
    read_bytea,
    read_date,
    read_float4,
    read_float8,
    read_int2,
    read_int4,
    read_int8,
    read_interval,
    read_json,
    read_line,
    read_lseg,
    read_macaddr,
    read_money,
    read_network,
    read_numeric,
    read_oid,
    read_path,
    read_point,
    read_polygon,
    read_serial2,
    read_serial4,
    read_serial8,
    read_text,
    read_time,
    read_timestamp,
    read_timestamptz as read_stamptz,
    read_timetz,
    read_uuid,
    write_array,
    write_bits,
    write_bool,
    write_box,
    write_bytea,
    write_date,
    write_float4,
    write_float8,
    write_int2,
    write_int4,
    write_int8,
    write_interval,
    write_json,
    write_line,
    write_lseg,
    write_macaddr,
    write_money,
    write_network,
    write_numeric,
    write_oid,
    write_path,
    write_point,
    write_polygon,
    write_serial2,
    write_serial4,
    write_serial8,
    write_text,
    write_time,
    write_timestamp,
    write_timestamptz as write_stamptz,
    write_timetz,
    write_uuid,
)


class PGTypeFunc(NamedTuple):
    """Class for associate read and write functions."""

    name: str
    pytype: type
    length: int
    read: FunctionType
    write: FunctionType


class PostgreSQLDtype(PGTypeFunc, Enum):
    """Associate read and write functions with postgres data type."""

    Array = PGTypeFunc("Array", list, -1, read_array, write_array)
    Bit = PGTypeFunc("Bit", str, -1, read_bits, write_bits)
    Bool = PGTypeFunc("Bool", bool, 1, read_bool, write_bool)
    Box = PGTypeFunc("Box", tuple, 32, read_box, write_box)
    Bytes = PGTypeFunc("Bytes", bytes, -1, read_bytea, write_bytea)
    Cidr = PGTypeFunc("Cidr", IP4N | IP6N, -1, read_network, write_network)
    Circle = PGTypeFunc("Circle", tuple, 24, read_line, write_line)
    Date = PGTypeFunc("Date", date, 4, read_date, write_date)
    Float4 = PGTypeFunc("Float4", float, 4, read_float4, write_float4)
    Float8 = PGTypeFunc("Float8", float, 8, read_float8, write_float8)
    Inet = PGTypeFunc("Inet", IP4A | IP6A, -1, read_network, write_network)
    Int2 = PGTypeFunc("Int2", int, 2, read_int2, write_int2)
    Int4 = PGTypeFunc("Int4", int, 4, read_int4, write_int4)
    Int8 = PGTypeFunc("Int8", int, 8, read_int8, write_int8)
    Interval = PGTypeFunc("Interval", rd, 16, read_interval, write_interval)
    Json = PGTypeFunc("Json", dict, -1, read_json, write_json)
    Line = PGTypeFunc("Line", tuple, 24, read_line, write_line)
    Lseg = PGTypeFunc("Lseg", list, 32, read_lseg, write_lseg)
    Macaddr8 = PGTypeFunc("Macaddr8", str, 8, read_macaddr, write_macaddr)
    Macaddr = PGTypeFunc("Macaddr", str, 6, read_macaddr, write_macaddr)
    Money = PGTypeFunc("Money", float, -1, read_money, write_money)
    Numeric = PGTypeFunc("Numeric", object, -1, read_numeric, write_numeric)
    Oid = PGTypeFunc("Oid", int, 4, read_oid, write_oid)
    Path = PGTypeFunc("Path", tuple | list, -1, read_path, write_path)
    Point = PGTypeFunc("Point", tuple, 16, read_point, write_point)
    Polygon = PGTypeFunc("Polygon", tuple, -1, read_polygon, write_polygon)
    Serial2 = PGTypeFunc("Serial2", int, 2, read_serial2, write_serial2)
    Serial4 = PGTypeFunc("Serial4", int, 4, read_serial4, write_serial4)
    Serial8 = PGTypeFunc("Serial8", int, 8, read_serial8, write_serial8)
    Text = PGTypeFunc("Text", str, -1, read_text, write_text)
    Time = PGTypeFunc("Time", time, 8, read_time, write_time)
    Timestamp = PGTypeFunc("Timestamp", dt, 8, read_timestamp, write_timestamp)
    Timestamptz = PGTypeFunc("Timestamptz", dt, 8, read_stamptz, write_stamptz)
    Timetz = PGTypeFunc("Timetz", time, 12, read_timetz, write_timetz)
    Uuid = PGTypeFunc("Uuid", object, 16, read_uuid, write_uuid)



from enum import Enum

from .dtypes import PostgreSQLDtype


class PGOid(Enum):
    """PGCopy OID Identifiers."""

    _bit = 1561
    _bool = 1000
    _box = 1020
    _bpchar = 1014
    _bytea = 1001
    _char = 1002
    _cidr = 651
    _circle = 719
    _date = 1182
    _float4 = 1021
    _float8 = 1022
    _inet = 1041
    _int2 = 1005
    _int4 = 1007
    _int8 = 1016
    _interval = 1187
    _json = 199
    _jsonb = 3807
    _line = 629
    _lseg = 1018
    _macaddr = 1040
    _macaddr8 = 775
    _money = 791
    _numeric = 1231
    _oid = 1028
    _path = 1019
    _point = 1017
    _polygon = 1027
    _text = 1009
    _time = 1183
    _timestamp = 1115
    _timestamptz = 1185
    _timetz = 1270
    _tsquery = 3645
    _tsvector = 3643
    _uuid = 2951
    _varbit = 1563
    _varchar = 1015
    _xml = 143
    bit = 1560
    bool = 16
    box = 603
    bpchar = 1042
    bytea = 17
    char = 18
    cidr = 650
    circle = 718
    date = 1082
    float4 = 700
    float8 = 701
    inet = 869
    int2 = 21
    int4 = 23
    int8 = 20
    interval = 1186
    json = 114
    jsonb = 3802
    line = 628
    lseg = 601
    macaddr = 829
    macaddr8 = 774
    money = 790
    numeric = 1700
    oid = 26
    path = 602
    point = 600
    polygon = 604
    text = 25
    time = 1083
    timestamp = 1114
    timestamptz = 1184
    timetz = 1266
    tsquery = 3615
    tsvector = 3614
    uuid = 2950
    varbit = 1562
    varchar = 1043
    xml = 142


ArrayOidToOid: dict[PGOid, PGOid] = {
    PGOid._bit: PGOid.bit,
    PGOid._bool: PGOid.bool,
    PGOid._box: PGOid.box,
    PGOid._bpchar: PGOid.bpchar,
    PGOid._bytea: PGOid.bytea,
    PGOid._char: PGOid.char,
    PGOid._cidr: PGOid.cidr,
    PGOid._circle: PGOid.circle,
    PGOid._date: PGOid.date,
    PGOid._float4: PGOid.float4,
    PGOid._float8: PGOid.float8,
    PGOid._inet: PGOid.inet,
    PGOid._int2: PGOid.int2,
    PGOid._int4: PGOid.int4,
    PGOid._int8: PGOid.int8,
    PGOid._interval: PGOid.interval,
    PGOid._json: PGOid.json,
    PGOid._jsonb: PGOid.jsonb,
    PGOid._line: PGOid.line,
    PGOid._lseg: PGOid.lseg,
    PGOid._macaddr: PGOid.macaddr,
    PGOid._macaddr8: PGOid.macaddr8,
    PGOid._money: PGOid.money,
    PGOid._numeric: PGOid.numeric,
    PGOid._oid: PGOid.oid,
    PGOid._path: PGOid.path,
    PGOid._point: PGOid.point,
    PGOid._polygon: PGOid.polygon,
    PGOid._text: PGOid.text,
    PGOid._time: PGOid.time,
    PGOid._timestamp: PGOid.timestamp,
    PGOid._timestamptz: PGOid.timestamptz,
    PGOid._timetz: PGOid.timetz,
    PGOid._uuid: PGOid.uuid,
    PGOid._varbit: PGOid.varbit,
    PGOid._varchar: PGOid.varchar,
    PGOid._xml: PGOid.xml,
}


PGOidToDType: dict[PGOid, PostgreSQLDtype] = {
    # Associate oid with data type.
    PGOid._bit: PostgreSQLDtype.Array,
    PGOid._bool: PostgreSQLDtype.Array,
    PGOid._box: PostgreSQLDtype.Array,
    PGOid._bpchar: PostgreSQLDtype.Array,
    PGOid._bytea: PostgreSQLDtype.Array,
    PGOid._char: PostgreSQLDtype.Array,
    PGOid._cidr: PostgreSQLDtype.Array,
    PGOid._circle: PostgreSQLDtype.Array,
    PGOid._date: PostgreSQLDtype.Array,
    PGOid._float4: PostgreSQLDtype.Array,
    PGOid._float8: PostgreSQLDtype.Array,
    PGOid._inet: PostgreSQLDtype.Array,
    PGOid._int2: PostgreSQLDtype.Array,
    PGOid._int4: PostgreSQLDtype.Array,
    PGOid._int8: PostgreSQLDtype.Array,
    PGOid._interval: PostgreSQLDtype.Array,
    PGOid._json: PostgreSQLDtype.Array,
    PGOid._jsonb: PostgreSQLDtype.Array,
    PGOid._line: PostgreSQLDtype.Array,
    PGOid._lseg: PostgreSQLDtype.Array,
    PGOid._macaddr: PostgreSQLDtype.Array,
    PGOid._macaddr8: PostgreSQLDtype.Array,
    PGOid._money: PostgreSQLDtype.Array,
    PGOid._numeric: PostgreSQLDtype.Array,
    PGOid._oid: PostgreSQLDtype.Array,
    PGOid._path: PostgreSQLDtype.Array,
    PGOid._point: PostgreSQLDtype.Array,
    PGOid._polygon: PostgreSQLDtype.Array,
    PGOid._text: PostgreSQLDtype.Array,
    PGOid._time: PostgreSQLDtype.Array,
    PGOid._timestamp: PostgreSQLDtype.Array,
    PGOid._timestamptz: PostgreSQLDtype.Array,
    PGOid._timetz: PostgreSQLDtype.Array,
    PGOid._uuid: PostgreSQLDtype.Array,
    PGOid._varbit: PostgreSQLDtype.Array,
    PGOid._varchar: PostgreSQLDtype.Array,
    PGOid._xml: PostgreSQLDtype.Array,
    PGOid.bit: PostgreSQLDtype.Bit,
    PGOid.bool: PostgreSQLDtype.Bool,
    PGOid.box: PostgreSQLDtype.Box,
    PGOid.bpchar: PostgreSQLDtype.Text,
    PGOid.bytea: PostgreSQLDtype.Bytes,
    PGOid.char: PostgreSQLDtype.Text,
    PGOid.cidr: PostgreSQLDtype.Cidr,
    PGOid.circle: PostgreSQLDtype.Circle,
    PGOid.date: PostgreSQLDtype.Date,
    PGOid.float4: PostgreSQLDtype.Float4,
    PGOid.float8: PostgreSQLDtype.Float8,
    PGOid.inet: PostgreSQLDtype.Inet,
    PGOid.int2: PostgreSQLDtype.Int2,
    PGOid.int4: PostgreSQLDtype.Int4,
    PGOid.int8: PostgreSQLDtype.Int8,
    PGOid.interval: PostgreSQLDtype.Interval,
    PGOid.json: PostgreSQLDtype.Json,
    PGOid.jsonb: PostgreSQLDtype.Json,
    PGOid.line: PostgreSQLDtype.Line,
    PGOid.lseg: PostgreSQLDtype.Lseg,
    PGOid.macaddr: PostgreSQLDtype.Macaddr,
    PGOid.macaddr8: PostgreSQLDtype.Macaddr8,
    PGOid.money: PostgreSQLDtype.Money,
    PGOid.numeric: PostgreSQLDtype.Numeric,
    PGOid.oid: PostgreSQLDtype.Oid,
    PGOid.path: PostgreSQLDtype.Path,
    PGOid.point: PostgreSQLDtype.Point,
    PGOid.polygon: PostgreSQLDtype.Polygon,
    PGOid.text: PostgreSQLDtype.Text,
    PGOid.time: PostgreSQLDtype.Time,
    PGOid.timestamp: PostgreSQLDtype.Timestamp,
    PGOid.timestamptz: PostgreSQLDtype.Timestamptz,
    PGOid.timetz: PostgreSQLDtype.Timetz,
    PGOid.uuid: PostgreSQLDtype.Uuid,
    PGOid.varbit: PostgreSQLDtype.Bit,
    PGOid.varchar: PostgreSQLDtype.Text,
    PGOid.xml: PostgreSQLDtype.Text,
}



from collections.abc import Generator
from io import BufferedReader
from typing import Any


def read_num_columns(
    fileobj: BufferedReader,
    column_length: int,
) -> int:
    """Read column count."""

    ...


def read_record(
    fileobj: BufferedReader,
    reader: object,
    pgoid_function: object,
    buffer_object: object,
    pgoid: int,
) -> Any:
    """Read one record."""

    ...


def skip_all(
    fileobj: BufferedReader,
    column_length: int,
    num_columns: int,
    num_rows: int,
) -> int:
    """Skip all records."""

    ...


def writer(
    fileobj: BufferedReader,
    write_row: object,
    dtype_values: Any,
    num_columns: int,
) -> int:
    """Write pgcopy into fileobj."""

    ...


def nullable_writer(
    write_dtype: object,
    dtype_value: Any,
    pgoid_function: object,
    buffer_object: object,
    pgoid: int,
) -> bytes:
    """Function for write None value and data with length."""

    ...


def make_rows(
    write_row: object,
    dtype_values: Any,
    num_columns: int,
) -> Generator[bytes, None, None]:
    """Make pgcopy rows."""

    ...



cpdef long read_num_columns(
    object fileobj,
    long column_length,
)
cpdef object read_record(
    object fileobj,
    object reader,
    object pgoid_function,
    object buffer_object,
    long pgoid,
)
cpdef long long skip_all(
    object fileobj,
    long column_length,
    long num_columns,
    long long num_rows,
)
cdef void skip_record(object fileobj)
cpdef long long writer(
    object fileobj,
    object write_row,
    object dtype_values,
    long num_columns,
)
cpdef bytes nullable_writer(
    object write_dtype,
    object dtype_value,
    object pgoid_function,
    object buffer_object,
    long pgoid,
)



cpdef str read_text(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack text value."""

    return binary_data.decode("utf-8", errors="replace")


cpdef bytes write_text(
    str dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack text value."""

    return str(dtype_value).encode("utf-8")


cpdef str read_macaddr(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack macaddr and macaddr8 value."""

    cdef long i
    cdef long data_len = len(binary_data)
    cdef const unsigned char[:] view = binary_data
    cdef list parts = []

    parts.reserve(data_len)

    for i in range(data_len):
        parts.append(f"{view[i]:02x}")

    return ":".join(parts).upper()


cpdef bytes write_macaddr(
    str dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack macaddr and macaddr8 value."""

    return bytes.fromhex(dtype_value.replace(":", ""))


cpdef str read_bits(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack bit and varbit value."""

    cdef unsigned int length
    cdef const unsigned char[:] view = binary_data
    cdef long data_len = len(binary_data)
    cdef long bit_data_len = data_len - 4
    cdef list bits = []
    cdef short i, j
    cdef unsigned char byte_val

    length = (view[0] << 24) | (view[1] << 16) | (view[2] << 8) | view[3]
    bits.reserve(bit_data_len * 8)

    for i in range(4, data_len):
        byte_val = view[i]
        for j in range(7, -1, -1):
            bits.append(str((byte_val >> j) & 1))

    return "".join(bits)[:length]


cpdef bytes write_bits(
    str dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack bit and varbit value."""

    cdef long bit_length = len(dtype_value)
    cdef long byte_length = (bit_length + 7) // 8
    cdef int int_value = int(dtype_value, 2)

    return int_value.to_bytes(byte_length, "big")


cpdef bytes read_bytea(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack bytea value."""

    return binary_data


cpdef bytes write_bytea(
    bytes dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack bytea value."""

    return dtype_value



from struct import (
    pack,
    unpack,
)


cpdef (double, double) read_point(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack point value."""

    return unpack("!2d", binary_data)


cpdef bytes write_point(
    (double, double) dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack point value."""

    return pack("!2d", *dtype_value)


cpdef (double, double, double) read_line(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack line value."""

    return unpack("!3d", binary_data)


cpdef bytes write_line(
    (double, double, double) dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack line value."""

    return pack("!3d", *dtype_value)


cpdef list read_lseg(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack lseg value."""

    cdef double x1, y1, x2, y2
    x1, y1, x2, y2 = unpack("!4d", binary_data)
    return [(x1, y1), (x2, y2)]


cpdef bytes write_lseg(
    list dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack lseg value."""

    return pack("!4d", *dtype_value[0], *dtype_value[1])


cpdef ((double, double), (double, double)) read_box(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack box value."""

    cdef double x1, y1, x2, y2
    x1, y1, x2, y2 = unpack("!4d", binary_data)
    return (x1, y1), (x2, y2)


cpdef bytes write_box(
    ((double, double), (double, double)) dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack box value."""

    return pack("!4d", *dtype_value[0], *dtype_value[1])


cpdef object read_path(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack path value."""

    cdef bint is_closed
    cdef int length
    cdef tuple coords_data
    cdef list path_data = []
    cdef long i

    is_closed = unpack("!?", binary_data[:1])[0]
    length = unpack("!l", binary_data[1:5])[0]

    cdef long coords_count = (len(binary_data) - 5) // 8
    coords_data = unpack(f"!{coords_count}d", binary_data[5:])

    for i in range(0, coords_count, 2):
        if i + 1 < coords_count:
            path_data.append((coords_data[i], coords_data[i + 1]))

    if is_closed:
        return tuple(path_data)
    return path_data


cpdef bytes write_path(
    object dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack path value."""

    cdef bint is_closed = isinstance(dtype_value, tuple)
    cdef short length = len(dtype_value)
    cdef list path_data = []
    cdef short i
    cdef object point

    for i in range(length):
        point = dtype_value[i]
        path_data.append(point[0])
        path_data.append(point[1])

    return pack(
        f"?l{len(path_data)}d",
        is_closed,
        length,
        *path_data
    )


cpdef tuple read_polygon(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack polygon value."""

    cdef int length
    cdef tuple coords_data
    cdef list points = []
    cdef short i
    length = unpack("!l", binary_data[:4])[0]
    cdef short coords_count = (len(binary_data) - 4) // 8
    coords_data = unpack(f"!{coords_count}d", binary_data[4:])

    for i in range(0, coords_count, 2):
        points.append((coords_data[i], coords_data[i + 1]))

    return tuple(points)


cpdef bytes write_polygon(
    tuple dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack polygon value."""

    cdef short length = len(dtype_value)
    cdef list path_data = []
    cdef short i
    cdef tuple point

    for i in range(length):
        point = dtype_value[i]
        path_data.extend(point)

    return pack(
        f"l{len(path_data)}d",
        length,
        *path_data
    )



from json import (
    dumps,
    loads,
)


cpdef object read_json(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack json value."""

    return loads(binary_data)


cpdef bytes write_json(
    object dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack json value."""

    return dumps(dtype_value, ensure_ascii=False).encode("utf-8")



from datetime import (
    date,
    datetime,
    time,
)
from dateutil.relativedelta import relativedelta


def read_date(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> date:
    """Unpack date value."""

    ...


def write_date(
    dtype_value: date,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack date value."""

    ...


def read_timestamp(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> datetime:
    """Unpack timestamp value."""

    ...


def write_timestamp(
    dtype_value: datetime,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack timestamp value."""

    ...


def read_timestamptz(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> datetime:
    """Unpack timestamptz value."""

    ...


def write_timestamptz(
    dtype_value: datetime,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack timestamptz value."""

    ...


def read_time(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> time:
    """Unpack time value."""

    ...


def write_time(
    dtype_value: time,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack time value."""

    ...


def read_timetz(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> time:
    """Unpack timetz value."""

    ...


def write_timetz(
    dtype_value: time,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack timetz value."""

    ...


def read_interval(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> relativedelta:
    """Unpack interval value."""

    ...


def write_interval(
    dtype_value: relativedelta,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack interval value."""

    ...



"""PGCopy bynary dump parser."""

from .common import (
    PGCopyRecordError,
    PGCopySignatureError,
    PGOid,
)
from .reader import PGCopyReader
from .writer import PGCopyWriter


__all__ = (
    "PGCopyReader",
    "PGCopyRecordError",
    "PGCopySignatureError",
    "PGCopyWriter",
    "PGOid",
)
__author__ = "0xMihalich"
__version__ = "0.2.1.8"



cpdef object read_bool(
    bytes binary_data,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes write_bool(
    object dtype_value,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef unsigned long read_oid(
    bytes binary_data,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes write_oid(
    object dtype_value,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef unsigned short read_serial2(
    bytes binary_data,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes write_serial2(
    object dtype_value,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef unsigned long read_serial4(
    bytes binary_data,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes write_serial4(
    object dtype_value,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef unsigned long long read_serial8(
    bytes binary_data,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes write_serial8(
    object dtype_value,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef short read_int2(
    bytes binary_data,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes write_int2(
    object dtype_value,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef long read_int4(
    bytes binary_data,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes write_int4(
    object dtype_value,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef long long read_int8(
    bytes binary_data,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes write_int8(
    object dtype_value,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef double read_money(
    bytes binary_data,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes write_money(
    object dtype_value,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef float read_float4(
    bytes binary_data,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes write_float4(
    object dtype_value,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef double read_float8(
    bytes binary_data,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes write_float8(
    object dtype_value,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef object read_numeric(
    bytes binary_data,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes write_numeric(
    object dtype_value,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)



from ipaddress import (
    IPv4Address,
    IPv4Network,
    IPv6Address,
    IPv6Network,
)
from struct import (
    pack,
    unpack,
)


cdef dict IpAddr = {
    2: IPv4Address,
    3: IPv6Address,
    IPv4Address: 2,
    IPv4Network: 2,
    IPv6Network: 3,
    IPv6Address: 3,
}
cdef dict IpNet = {
    IPv4Address: IPv4Network,
    IPv6Address: IPv6Network,
}


cpdef object read_network(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack inet or cidr value."""

    cdef int binary_length = len(binary_data) - 4
    cdef unsigned char ip_family, ip_netmask, is_cidr, ip_length
    cdef bytes ip_data
    ip_family, ip_netmask, is_cidr, ip_length, ip_data = unpack(
        f"!4B{binary_length}s",
        binary_data,
    )

    cdef object ip_addr = IpAddr[ip_family](ip_data)

    if is_cidr:
        return IpNet[ip_addr.__class__](
            f"{ip_addr}/{ip_netmask}",
            strict=False,
        )

    return ip_addr


cpdef bytes write_network(
    object dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack inet or cidr value."""

    cdef unsigned char ip_family, ip_netmask, is_cidr, ip_length
    cdef bytes ip_addr

    if isinstance(dtype_value, IPv4Address | IPv6Address):
        ip_addr = dtype_value.packed
        ip_netmask = dtype_value.max_prefixlen
        is_cidr = 0
    elif isinstance(dtype_value, IPv4Network | IPv6Network):
        ip_addr = dtype_value.network_address.packed
        ip_netmask = dtype_value._prefixlen
        is_cidr = 1

    ip_family = IpAddr[dtype_value]
    ip_length = len(ip_addr)

    return pack(
        f"!4B{ip_length}s",
        ip_family,
        ip_netmask,
        is_cidr,
        ip_length,
        ip_addr,
    )



from uuid import UUID


def read_uuid(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> UUID:
    """Unpack uuid value."""

    ...


def write_uuid(
    dtype_value: UUID,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack uuid value."""

    ...



cpdef object read_date(
    bytes binary_data,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes write_date(
    object dtype_value,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef object read_timestamp(
    bytes binary_data,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes write_timestamp(
    object dtype_value,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef object read_timestamptz(
    bytes binary_data,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes write_timestamptz(
    object dtype_value,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef object read_time(
    bytes binary_data,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes write_time(
    object dtype_value,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef object read_timetz(
    bytes binary_data,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes write_timetz(
    object dtype_value,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef object read_interval(
    bytes binary_data,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes write_interval(
    object dtype_value,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)



def read_json(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> dict:
    """Unpack json value."""

    ...


def write_json(
    dtype_value: dict,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack json value."""

    ...



from datetime import (
    date,
    datetime,
    time,
    timedelta,
    timezone,
)
from dateutil.relativedelta import relativedelta
from struct import (
    pack,
    unpack,
    unpack_from,
)


cdef object DEFAULT_DATE = date(2000, 1, 1)
cdef object DEFAULT_DATETIME = datetime(2000, 1, 1)
cdef long long MICROSECONDS_PER_SECOND = 1_000_000
cdef int SECONDS_PER_MINUTE = 60
cdef int HOURS_PER_DAY = 24
cdef int SECONDS_PER_HOUR = 3600


cpdef object read_date(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack date value."""

    return DEFAULT_DATE + timedelta(days=unpack("!i", binary_data)[0])


cpdef bytes write_date(
    object dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack date value."""

    if dtype_value.__class__.__name__ in ("Timestamp", "datetime"):
        dtype_value = dtype_value.date()

    cdef int days = (dtype_value - DEFAULT_DATE).days
    return pack("!i", days)


cpdef object read_timestamp(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack timestamp value."""

    return DEFAULT_DATETIME + timedelta(
        microseconds=unpack("!q", binary_data)[0]
    )


cpdef bytes write_timestamp(
    object dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack timestamp value."""

    if dtype_value.__class__.__name__ == "Timestamp":
        dtype_value = dtype_value.to_pydatetime()
    elif dtype_value.__class__ is date:
        dtype_value = datetime.combine(dtype_value, datetime.min.time())

    cdef long long seconds = (dtype_value - DEFAULT_DATETIME).total_seconds()
    cdef long long microseconds = seconds * MICROSECONDS_PER_SECOND
    return pack("!q", microseconds)


cpdef object read_timestamptz(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack timestamptz value."""

    return read_timestamp(binary_data).replace(tzinfo=timezone.utc)


cpdef bytes write_timestamptz(
    object dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack timestamptz value."""

    return write_timestamp(dtype_value.astimezone(timezone.utc))


cpdef object read_time(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack time value."""

    cdef long long microseconds, total_seconds, total_minutes
    cdef long long hours, minutes, seconds, microsecond

    microseconds = unpack_from("!q", binary_data)[0]
    total_seconds = microseconds // MICROSECONDS_PER_SECOND
    microsecond = microseconds % MICROSECONDS_PER_SECOND
    seconds = total_seconds % SECONDS_PER_MINUTE
    total_minutes = total_seconds // SECONDS_PER_MINUTE
    minutes = total_minutes % SECONDS_PER_MINUTE
    hours = total_minutes // SECONDS_PER_MINUTE
    hours = hours % HOURS_PER_DAY

    return time(hours, minutes, seconds, microsecond)


cpdef bytes write_time(
    object dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack time value."""

    cdef long long hour_per_microsecond = (
        dtype_value.hour * SECONDS_PER_HOUR * MICROSECONDS_PER_SECOND
    )
    cdef long long minute_per_microsecond = (
        dtype_value.minute * SECONDS_PER_MINUTE * MICROSECONDS_PER_SECOND
    )
    cdef long long second_per_microsecond = (
        dtype_value.second * MICROSECONDS_PER_SECOND
    )
    cdef long long total_microseconds = (
        hour_per_microsecond +
        minute_per_microsecond +
        second_per_microsecond +
        dtype_value.microsecond
    )
    return pack("!q", total_microseconds)


cpdef object read_timetz(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack timetz value."""

    cdef object time_notz = read_time(binary_data)
    cdef long tz_offset_sec = unpack("!i", binary_data[-4:])[0]
    cdef object tz_offset = timedelta(seconds=tz_offset_sec)
    return time_notz.replace(tzinfo=timezone(tz_offset))


cpdef bytes write_timetz(
    object dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack timetz value."""

    cdef long long hour_per_microsecond = (
        dtype_value.hour * SECONDS_PER_HOUR * MICROSECONDS_PER_SECOND
    )
    cdef long long minute_per_microsecond = (
        dtype_value.minute * SECONDS_PER_MINUTE * MICROSECONDS_PER_SECOND
    )
    cdef long long second_per_microsecond = (
        dtype_value.second * MICROSECONDS_PER_SECOND
    )
    cdef long long total_microseconds = (
        hour_per_microsecond +
        minute_per_microsecond +
        second_per_microsecond +
        dtype_value.microsecond
    )
    cdef object tz_offset = dtype_value.tzinfo.utcoffset(None)
    cdef int offset = tz_offset.total_seconds()
    return pack("!qi", total_microseconds, offset)


cpdef object read_interval(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack interval value."""

    cdef long long microseconds
    cdef int days, months
    microseconds, days, months = unpack("!qii", binary_data)
    return relativedelta(months=months, days=days, microseconds=microseconds)


cpdef bytes write_interval(
    object dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack interval value."""

    cdef long long days = dtype_value.days or 0
    cdef long long months = dtype_value.months or 0
    cdef long long hours = dtype_value.hours or 0
    cdef long long minutes = dtype_value.minutes or 0
    cdef long long seconds = dtype_value.seconds or 0
    cdef long long microseconds = dtype_value.microseconds or 0
    cdef long long hour_per_microsecond = (
        hours * SECONDS_PER_HOUR * MICROSECONDS_PER_SECOND
    )
    cdef long long minute_per_microsecond = (
        minutes * SECONDS_PER_MINUTE * MICROSECONDS_PER_SECOND
    )
    cdef long long second_per_microsecond = (
        seconds * MICROSECONDS_PER_SECOND
    )
    cdef long long total_microseconds = (
        hour_per_microsecond +
        minute_per_microsecond +
        second_per_microsecond +
        microseconds
    )
    return pack("!qii", total_microseconds, days, months)



from uuid import UUID


cpdef object read_uuid(
    bytes binary_data,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Unpack uuid value."""

    return UUID(bytes=binary_data)


cpdef bytes write_uuid(
    object dtype_value,
    object pgoid_function = None,
    object buffer_object = None,
    object pgoid = None,
):
    """Pack uuid value."""

    return dtype_value.bytes



from cpython cimport PyBytes_AsString
from struct import (
    pack,
    unpack,
)


cdef bytes NULLABLE = b"\xff\xff\xff\xff"


cdef list recursive_elements(list elements, list array_struct):
    """Recursive unpack array struct."""

    cdef long chunk, num_chunks
    cdef long _i, start, end
    cdef long elements_len = len(elements)
    cdef list result = []

    if not array_struct:
        return elements

    chunk = array_struct.pop()

    if elements_len == chunk:
        return recursive_elements(elements, array_struct)
    
    num_chunks = (elements_len + chunk - 1) // chunk

    for _i in range(num_chunks):
        start = _i * chunk
        end = start + chunk

        if end > elements_len:
            end = elements_len
        result.append(elements[start:end])

    return recursive_elements(result, array_struct)


cdef list get_num_dim(object type_values):
    """Get list of num dim."""

    cdef list num_dim = []
    cdef object current = type_values

    while current.__class__ is list and len(current) > 0:
        num_dim.append(len(current))
        current = current[0]

    return num_dim


cdef long prod(list iterable):
    """Cython math.prod."""

    cdef long result, item

    for item in iterable:
        result *= item

    return result


cdef object _reader(object buffer_object, object pgoid_function):
    """Read array record."""

    cdef bytes _bytes = buffer_object.read(4)
    cdef const unsigned char *buf = <const unsigned char*>PyBytes_AsString(
        _bytes
    )
    cdef int length = (buf[0] << 24) | (buf[1] << 16) | (buf[2] << 8) | buf[3]

    if length == 0xffffffff:
        return

    return pgoid_function(buffer_object.read(length))


cpdef list read_array(
    bytes binary_data,
    object pgoid_function,
    object buffer_object,
    long pgoid,
):
    """Unpack array values."""

    cdef unsigned int num_dim, _, oid
    cdef list array_struct = []
    cdef list array_elements = []

    buffer_object.write(binary_data)
    buffer_object.seek(0)
    num_dim, _, oid = unpack("!3I", buffer_object.read(12))

    for _ in range(num_dim):
        array_struct.append(unpack("!2I", buffer_object.read(8))[0])

    for _ in range(prod(array_struct)):
        array_elements.append(_reader(buffer_object, pgoid_function))

    buffer_object.seek(0)
    buffer_object.truncate()
    return recursive_elements(array_elements, array_struct)


cpdef bytes write_array(
    list dtype_value,
    object pgoid_function,
    object buffer_object,
    long pgoid,
):
    """Pack array values."""

    cdef list num_dim = get_num_dim(dtype_value)
    cdef short is_nullable, dim, dim_length = len(num_dim)
    cdef list expand_values, dimensions = []
    cdef object value
    cdef short length_dimensions
    cdef bytes binary_data
    cdef bint has_list = True

    while has_list:
        has_list = False
        expand_values = []

        for value in dtype_value:
            if isinstance(value, list):
                has_list = True
                expand_values.extend(value)
            else:
                expand_values.append(value)

        dtype_value = expand_values

    is_nullable = False
    for value in dtype_value:
        if value is None:
            is_nullable = True
            break

    for dim in num_dim:
        dimensions.extend([dim, 1])

    length_dimensions = len(dimensions)

    buffer_object.write(pack("!3I", dim_length, is_nullable, pgoid))
    buffer_object.write(pack("!%dI" % length_dimensions, *dimensions))

    for value in dtype_value:
        if value is None:
            buffer_object.write(NULLABLE)
        else:
            binary_data = pgoid_function(value)
            buffer_object.write(pack("!I", len(binary_data)))
            buffer_object.write(binary_data)

    binary_data = buffer_object.getvalue()
    buffer_object.seek(0)
    buffer_object.truncate()
    return binary_data



cpdef (double, double) read_point(
    bytes binary_data,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes write_point(
    (double, double) dtype_value,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef (double, double, double) read_line(
    bytes binary_data,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes write_line(
    (double, double, double) dtype_value,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef list read_lseg(
    bytes binary_data,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes write_lseg(
    list dtype_value,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef ((double, double), (double, double)) read_box(
    bytes binary_data,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes write_box(
    ((double, double), (double, double)) dtype_value,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef object read_path(
    bytes binary_data,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes write_path(
    object dtype_value,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef tuple read_polygon(
    bytes binary_data,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes write_polygon(
    tuple dtype_value,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)



cpdef object read_json(
    bytes binary_data,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)
cpdef bytes write_json(
    object dtype_value,
    object pgoid_function=*,
    object buffer_object=*,
    object pgoid=*,
)



def read_text(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> str:
    """Unpack text value."""

    ...


def write_text(
    dtype_value: str,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack text value."""

    ...


def read_macaddr(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> str:
    """Unpack macaddr and macaddr8 value."""

    ...


def write_macaddr(
    dtype_value: str,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack macaddr and macaddr8 value."""

    ...


def read_bits(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> str:
    """Unpack bit and varbit value."""

    ...


def write_bits(
    dtype_value: str,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack bit and varbit value."""

    ...


def read_bytea(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Unpack bytea value."""

    ...


def write_bytea(
    dtype_value: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack bytea value."""

    ...



from ipaddress import (
    IPv4Address,
    IPv4Network,
    IPv6Address,
    IPv6Network,
)


def read_network(
    binary_data: bytes,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> IPv4Address | IPv4Network | IPv6Address | IPv6Network:
    """Unpack inet or cidr value."""

    ...


def write_network(
    dtype_value: IPv4Address | IPv4Network | IPv6Address | IPv6Network,
    pgoid_function: object | None = None,
    buffer_object: object | None = None,
    pgoid: int | None = None,
) -> bytes:
    """Pack inet or cidr value."""

    ...



from setuptools import (
    Extension,
    setup,
)
from Cython.Build import cythonize

extensions = [
    Extension(
        "pgcopylib.common.base",
        ["src/pgcopylib/common/base.pyx"],
    ),
    Extension(
        "pgcopylib.common.dtypes.functions.arrays",
        ["src/pgcopylib/common/dtypes/functions/arrays.pyx"],
    ),
    Extension(
        "pgcopylib.common.dtypes.functions.dates",
        ["src/pgcopylib/common/dtypes/functions/dates.pyx"],
    ),
    Extension(
        "pgcopylib.common.dtypes.functions.digits",
        ["src/pgcopylib/common/dtypes/functions/digits.pyx"],
    ),
    Extension(
        "pgcopylib.common.dtypes.functions.geometrics",
        ["src/pgcopylib/common/dtypes/functions/geometrics.pyx"],
    ),
    Extension(
        "pgcopylib.common.dtypes.functions.ipaddrs",
        ["src/pgcopylib/common/dtypes/functions/ipaddrs.pyx"],
    ),
    Extension(
        "pgcopylib.common.dtypes.functions.jsons",
        ["src/pgcopylib/common/dtypes/functions/jsons.pyx"],
    ),
    Extension(
        "pgcopylib.common.dtypes.functions.strings",
        ["src/pgcopylib/common/dtypes/functions/strings.pyx"],
    ),
    Extension(
        "pgcopylib.common.dtypes.functions.uuids",
        ["src/pgcopylib/common/dtypes/functions/uuids.pyx"],
    ),
]

setup(
    name="pgcopylib",
    package_dir={"": "src"},
    ext_modules=cythonize(extensions, language_level="3"),
    packages=[
        "pgcopylib.common",
        "pgcopylib.common.dtypes.functions",
    ],
    package_data={
        "pgcopylib": [
            "**/*.pyx",
            "**/*.pyi",
            "**/*.pxd",
            "*.pxd",
            "*.pyd",
            "*.md",
            "*.txt",
        ]
    },
    exclude_package_data={
        "": ["*.c"],
        "pgcopylib": ["**/*.c"],
    },
    include_package_data=True,
    setup_requires=["Cython>=3.0"],
    zip_safe=False,
)



"""Convert between bytes and data types functions."""

from .arrays import (
    read_array,
    write_array,
)
from .dates import (
    read_date,
    read_interval,
    read_time,
    read_timestamp,
    read_timestamptz,
    read_timetz,
    write_date,
    write_interval,
    write_time,
    write_timestamp,
    write_timestamptz,
    write_timetz,
)
from .digits import (
    read_bool,
    read_float4,
    read_float8,
    read_int2,
    read_int4,
    read_int8,
    read_money,
    read_numeric,
    read_oid,
    read_serial2,
    read_serial4,
    read_serial8,
    write_bool,
    write_float4,
    write_float8,
    write_int2,
    write_int4,
    write_int8,
    write_money,
    write_numeric,
    write_oid,
    write_serial2,
    write_serial4,
    write_serial8,
)
from .geometrics import (
    read_box,
    read_line,
    read_lseg,
    read_path,
    read_point,
    read_polygon,
    write_box,
    write_line,
    write_lseg,
    write_path,
    write_point,
    write_polygon,
)
from .ipaddrs import (
    read_network,
    write_network,
)
from .jsons import (
    read_json,
    write_json,
)
from .strings import (
    read_bits,
    read_bytea,
    read_macaddr,
    read_text,
    write_bits,
    write_bytea,
    write_macaddr,
    write_text,
)
from .uuids import (
    read_uuid,
    write_uuid,
)


__all__ = (
    "read_array",
    "read_bits",
    "read_bool",
    "read_box",
    "read_bytea",
    "read_date",
    "read_float4",
    "read_float8",
    "read_int2",
    "read_int4",
    "read_int8",
    "read_interval",
    "read_json",
    "read_line",
    "read_lseg",
    "read_macaddr",
    "read_money",
    "read_network",
    "read_numeric",
    "read_oid",
    "read_path",
    "read_point",
    "read_polygon",
    "read_serial2",
    "read_serial4",
    "read_serial8",
    "read_text",
    "read_time",
    "read_timestamp",
    "read_timestamptz",
    "read_timetz",
    "read_uuid",
    "write_array",
    "write_bits",
    "write_bool",
    "write_box",
    "write_bytea",
    "write_date",
    "write_float4",
    "write_float8",
    "write_int2",
    "write_int4",
    "write_int8",
    "write_interval",
    "write_json",
    "write_line",
    "write_lseg",
    "write_macaddr",
    "write_money",
    "write_network",
    "write_numeric",
    "write_oid",
    "write_path",
    "write_point",
    "write_polygon",
    "write_serial2",
    "write_serial4",
    "write_serial8",
    "write_text",
    "write_time",
    "write_timestamp",
    "write_timestamptz",
    "write_timetz",
    "write_uuid",
)



from .enums import (
    ArrayOidToOid,
    PGOid,
    PGOidToDType,
)
from .errors import (
    PGCopyRecordError,
    PGCopySignatureError,
)


__all__ = (
    "ArrayOidToOid",
    "PGCopyRecordError",
    "PGCopySignatureError",
    "PGOid",
    "PGOidToDType",
)
